/***********************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/
/* $XConsortium: mipolygenZ.c,v 1.3 93/12/09 16:15:44 takamori Exp $ */
#include "X.h"
#include "gcstruct.h"
#include "miscanfillZ.h"
#include "mipolyZ.h"
#include "pixmap.h"
#include "ddpex.h"
#include "ddpex3.h"

extern void miZloadAET(), miZcomputeWAET(), miZFreeStorage();

/*
 *
 *     Written by Brian Kelleher;  Oct. 1985
 *
 *     Routine to fill a polygon.  Two fill rules are
 *     supported: frWINDING and frEVENODD.
 *
 *     See fillpoly.h for a complete description of the algorithm.
 */

Bool
miZFillGeneralPoly(zbs_ptr, dst, pgc, count, ptsIn)
    ddZBufferStruct *zbs_ptr;
    DrawablePtr dst;
    GCPtr	pgc;
    int		count;              /* number of points        */
    ddCoord3DS  *ptsIn;              /* the points              */
{
    register ZEdgeTableEntry *pAET;  /* the Active Edge Table   */
    register int y;                 /* the current scanline    */
    register int nPts = 0;          /* number of pts in buffer */
    register ZEdgeTableEntry *pWETE; /* Winding Edge Table      */
    register ZScanLineList *pSLL;    /* Current ScanLineList    */
    register ddCoord3DS *ptsOut;    /* ptr to output buffers   */
    int *width;
    int *depth;
    ddCoord3DS FirstPoint[NUMPTSTOBUFFER]; /* the output buffers */
    int FirstWidth[NUMPTSTOBUFFER];
    int FirstDepth[NUMPTSTOBUFFER];
    ZEdgeTableEntry *pPrevAET;       /* previous AET entry      */
    ZEdgeTable ET;                   /* Edge Table header node  */
    ZEdgeTableEntry AET;             /* Active ET header node   */
    ZEdgeTableEntry *pETEs;          /* Edge Table Entries buff */
    ZScanLineListBlock SLLBlock;     /* header for ScanLineList */
    int fixWAET = 0;

    if (count < 3)
	return(TRUE);

    if(!(pETEs = (ZEdgeTableEntry *)
        ALLOCATE_LOCAL(sizeof(ZEdgeTableEntry) * count)))
	return(FALSE);
    ptsOut = FirstPoint;
    width = FirstWidth;
    depth = FirstDepth;
    if (!miZCreateETandAET(count, ptsIn, &ET, &AET, pETEs, &SLLBlock))
    {
	DEALLOCATE_LOCAL(pETEs);
	return(FALSE);
    }
    pSLL = ET.scanlines.next;
    if (pgc->fillRule == EvenOddRule) 
    {
        /*
         *  for each scanline
         */
        for (y = ET.ymin; y < ET.ymax; y++) 
        {
            /*
             *  Add a new edge to the active edge table when we
             *  get to the next edge.
             */
            if (pSLL && y == pSLL->scanline) 
            {
                miZloadAET(&AET, pSLL->edgelist);
                pSLL = pSLL->next;
            }
            pPrevAET = &AET;
            pAET = AET.next;

            /*
             *  for each active edge
             */
            while (pAET) 
            {
                ptsOut->x = pAET->bres.minor;
                ptsOut->z = pAET->bres.minorz;
		ptsOut++->y = y;
                *width++ = pAET->next->bres.minor - pAET->bres.minor;
                *depth++ = pAET->next->bres.minorz - pAET->bres.minorz;
                nPts++;

                /*
                 *  send out the buffer when its full
                 */
                if (nPts == NUMPTSTOBUFFER) 
		{
		    ZBuffer_FillSpans(zbs_ptr, dst, pgc,
				     nPts, FirstPoint, FirstWidth, FirstDepth);
                    ptsOut = FirstPoint;
                    width = FirstWidth;
                    depth = FirstDepth;
                    nPts = 0;
                }
                ZEVALUATEEDGEEVENODD(pAET, pPrevAET, y);
                ZEVALUATEEDGEEVENODD(pAET, pPrevAET, y);
            }
            miZInsertionSort(&AET);
        }
    }
    else      /* default to WindingNumber */
    {
        /*
         *  for each scanline
         */
        for (y = ET.ymin; y < ET.ymax; y++) 
        {
            /*
             *  Add a new edge to the active edge table when we
             *  get to the next edge.
             */
            if (pSLL && y == pSLL->scanline) 
            {
                miZloadAET(&AET, pSLL->edgelist);
                miZcomputeWAET(&AET);
                pSLL = pSLL->next;
            }
            pPrevAET = &AET;
            pAET = AET.next;
            pWETE = pAET;

            /*
             *  for each active edge
             */
            while (pAET) 
            {
                /*
                 *  if the next edge in the active edge table is
                 *  also the next edge in the winding active edge
                 *  table.
                 */
                if (pWETE == pAET) 
                {
                    ptsOut->x = pAET->bres.minor;
		    ptsOut++->y = y;
                    ptsOut->z = pAET->bres.minorz;
                    *width++ = pAET->nextWETE->bres.minor - pAET->bres.minor;
                    *depth++ = pAET->nextWETE->bres.minorz - pAET->bres.minorz;
                    nPts++;

                    /*
                     *  send out the buffer
                     */
                    if (nPts == NUMPTSTOBUFFER) 
                    {
		        ZBuffer_FillSpans(zbs_ptr, dst, pgc,
				     nPts, FirstPoint, FirstWidth, FirstDepth);
                        ptsOut = FirstPoint;
                        width  = FirstWidth;
                        depth  = FirstDepth;
                        nPts = 0;
                    }

                    pWETE = pWETE->nextWETE;
                    while (pWETE != pAET)
                        ZEVALUATEEDGEWINDING(pAET, pPrevAET, y, fixWAET);
                    pWETE = pWETE->nextWETE;
                }
                ZEVALUATEEDGEWINDING(pAET, pPrevAET, y, fixWAET);
            }

            /*
             *  reevaluate the Winding active edge table if we
             *  just had to resort it or if we just exited an edge.
             */
            if (miZInsertionSort(&AET) || fixWAET) 
            {
                miZcomputeWAET(&AET);
                fixWAET = 0;
            }
        }
    }

    /*
     *     Get any spans that we missed by buffering
     */
    ZBuffer_FillSpans(zbs_ptr, dst, pgc,
		      nPts, FirstPoint, FirstWidth, FirstDepth);
    DEALLOCATE_LOCAL(pETEs);
    miZFreeStorage(SLLBlock.next);
    return(TRUE);
}
