/*
 * $XConsortium: miwidelineZ.h,v 1.2 93/08/25 12:11:37 takamori Exp $
 *
 * Copyright 1988 Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * Author:  Keith Packard, MIT X Consortium
 */

#include "mispansZ.h"

/* 
 * interface data to span-merging polygon filler
 */

typedef struct _ZSpanData {
    ZSpanGroup	fgGroup, bgGroup;
} ZSpanDataRec, *ZSpanDataPtr;

extern ZSpanDataPtr  miSetupSpanDataZ ();

#define AppendSpanGroupZ(pGC, pixel, spanPtr, spanData) { \
	ZSpanGroup   *group, *othergroup = NULL; \
	if (pixel == pGC->fgPixel) \
	{ \
	    group = &spanData->fgGroup; \
	    if (pGC->lineStyle == LineDoubleDash) \
		othergroup = &spanData->bgGroup; \
	} \
	else \
	{ \
	    group = &spanData->bgGroup; \
	    othergroup = &spanData->fgGroup; \
	} \
	miAppendSpansZ (group, othergroup, spanPtr); \
}

/*
 * Polygon edge description for integer wide-line routines
 */

typedef struct _ZPolyEdge {
    int	    height;	/* number of scanlines to process */
    int	    x;		/* starting x coordinate */
    int	    z;		/* starting z coordinate */
    int	    stepx;	/* fixed integral dx */
    int	    stepz;	/* fixed integral dz */
    int	    signdx;	/* variable dx sign */
    int	    signdz;	/* variable dz sign */
    int	    e;		/* initial error term */
    int	    ez;		/* initial error term */
    int	    dz;
    int	    dy;
    int	    dx;
} ZPolyEdgeRec, *ZPolyEdgePtr;

#define SQSECANT 108.856472512142 /* 1/sin^2(11/2) - miter limit constant */

/*
 * types for general polygon routines
 */

typedef struct _ZPolyVertex {
    double  x, y, z;
} ZPolyVertexRec, *ZPolyVertexPtr;

typedef struct _ZPolySlope {
    int	    dx, dy, dz;
    double  k;	    /* x0 * dy - y0 * dx */
    double  kz;	    /* z0 * dy - y0 * dz */
} ZPolySlopeRec, *ZPolySlopePtr;

/*
 * Line face description for caps/joins
 */

typedef struct _ZLineFace {
    double  xa, ya, za;
    int	    dx, dy, dz;
    int	    x, y, z;
    double  k, kz;
} ZLineFaceRec, *ZLineFacePtr;

/*
 * macros for polygon fillers
 */

#define MIPOLYRELOADLEFTZ   if (!left_height && left_count) { \
	    	    	    	left_height = left->height; \
	    	    	    	left_x = left->x; \
	    	    	    	left_z = left->z; \
	    	    	    	left_stepx = left->stepx; \
	    	    	    	left_stepz = left->stepz; \
	    	    	    	left_signdx = left->signdx; \
	    	    	    	left_signdz = left->signdz; \
	    	    	    	left_e = left->e; \
	    	    	    	left_ez = left->ez; \
	    	    	    	left_dz = left->dz; \
	    	    	    	left_dy = left->dy; \
	    	    	    	left_dx = left->dx; \
	    	    	    	--left_count; \
	    	    	    	++left; \
			    }

#define MIPOLYRELOADRIGHTZ  if (!right_height && right_count) { \
	    	    	    	right_height = right->height; \
	    	    	    	right_x = right->x; \
	    	    	    	right_z = right->z; \
	    	    	    	right_stepx = right->stepx; \
	    	    	    	right_stepz = right->stepz; \
	    	    	    	right_signdx = right->signdx; \
	    	    	    	right_signdz = right->signdz; \
	    	    	    	right_e = right->e; \
	    	    	    	right_ez = right->ez; \
	    	    	    	right_dz = right->dz; \
	    	    	    	right_dy = right->dy; \
	    	    	    	right_dx = right->dx; \
	    	    	    	--right_count; \
	    	    	    	++right; \
			}

#define MIPOLYSTEPLEFTZ left_z += left_stepz; \
                        left_x += left_stepx; \
    	    	    	left_ez += left_dz; \
    	    	    	left_e += left_dx; \
    	    	    	if (left_e > 0) \
    	    	    	{ \
	    	    	    left_x += left_signdx; \
	    	    	    left_e -= left_dy; \
    	    	    	} \
    	    	    	if (left_ez > 0) \
    	    	    	{ \
	    	    	    left_z += left_signdz; \
	    	    	    left_ez -= left_dy; \
    	    	    	}

#define MIPOLYSTEPRIGHTZ right_x += right_stepx; \
    	    	    	right_z += right_stepz; \
    	    	    	right_e += right_dx; \
    	    	    	right_ez += right_dz; \
    	    	    	if (right_e > 0) \
    	    	    	{ \
	    	    	    right_x += right_signdx; \
	    	    	    right_e -= right_dy; \
    	    	    	} \
    	    	    	if (right_ez > 0) \
    	    	    	{ \
	    	    	    right_z += right_signdz; \
	    	    	    right_ez -= right_dy; \
    	    	    	}

#define MILINESETPIXEL(pDrawable, pGC, pixel, oldPixel) { \
    oldPixel = pGC->fgPixel; \
    if (pixel != oldPixel) { \
	DoChangeGC (pGC, GCForeground, (XID *) &pixel, FALSE); \
	ValidateGC (pDrawable, pGC); \
    } \
}
#define MILINERESETPIXEL(pDrawable, pGC, pixel, oldPixel) { \
    if (pixel != oldPixel) { \
	DoChangeGC (pGC, GCForeground, (XID *) &oldPixel, FALSE); \
	ValidateGC (pDrawable, pGC); \
    } \
}

#ifdef NOINLINEICEIL
#define ICEIL(x) ((int)ceil(x))
#else
#ifdef __GNUC__
static __inline int ICEIL(x)
    double x;
{
    int _cTmp = x;
    return ((x == _cTmp) || (x < 0.0)) ? _cTmp : _cTmp+1;
}
#else
#define ICEIL(x) ((((x) == (_cTmp = (x))) || ((x) < 0.0)) ? _cTmp : _cTmp+1)
#define ICEILTEMPDECL static int _cTmp;
#endif
#endif
