/***********************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/
/* $XConsortium: mizerlineZ.c,v 1.2 93/10/29 17:12:19 takamori Exp $ */
#include "X.h"

#include "misc.h"
#include "scrnintstr.h"
#include "gcstruct.h"
#include "windowstr.h"
#include "pixmap.h"
#include "miZ.h"
#include "ddpex3.h"

void
miZZeroLine(zbs_ptr, dst, pgc, mode, nptInit, pptInit)
ddZBufferStruct *zbs_ptr;
DrawablePtr dst;
GCPtr pgc;
int mode;
int nptInit;		/* number of points in polyline */
ddCoord3DS *pptInit;	/* points in the polyline */
{
    int xorg, yorg;
    ddCoord3DS *ppt;
    int npt;

    ddCoord3DS pt1, pt2;
    int dx, dy, dz;
    int adx, ady, adz;
    int signdx, signdy, signdz;
    register int len;

    int du, dv, dw;
    register int e, e1, e2, ez, ez1, ez2;

    register int x,y,z;	/* current point on the line */
    int i;		/* traditional name for loop counter */

    ddCoord3DS *pspanInit;
    int *pwidthInit, *pdepthInit;
    int list_len;
    Bool local = TRUE;

    list_len = (dst->height >= dst->width) ? dst->height : dst->width;
    pspanInit = (ddCoord3DS *)ALLOCATE_LOCAL(list_len * sizeof(ddCoord3DS));
    pwidthInit = (int *)ALLOCATE_LOCAL(list_len * sizeof(int));
    pdepthInit = (int *)ALLOCATE_LOCAL(list_len * sizeof(int));
    if (!pspanInit || !pwidthInit || !pdepthInit) 
    {
        if (pspanInit) DEALLOCATE_LOCAL(pspanInit);
        if (pwidthInit) DEALLOCATE_LOCAL(pwidthInit);
        if (pdepthInit) DEALLOCATE_LOCAL(pdepthInit);
	return;
    }
    ppt = pptInit;
    npt = nptInit;
    if (pgc->miTranslate)
    {
	xorg = dst->x;
	yorg = dst->y;

        if (mode == CoordModeOrigin) 
        {
	        for (i = 0; i<npt; i++) 
                {    
	            ppt->x += xorg;
	            ppt++->y += yorg;
	        }
        }
        else 
        {
	    ppt->x += xorg;
	    ppt++->y += yorg;
	    for (i = 1; i<npt; i++) 
            {
	        ppt->x += (ppt-1)->x;
	        ppt->y += (ppt-1)->y;
	        ppt++;
	    }
        }
    }
    else
    {
	if (mode == CoordModePrevious)
        {
	    ppt++;
	    for (i = 1; i<npt; i++) 
            {
	        ppt->x += (ppt-1)->x;
	        ppt->y += (ppt-1)->y;
	        ppt->z += (ppt-1)->z;
	        ppt++;
	    }
        }
    }

    ppt = pptInit;
    npt = nptInit;
    while (--npt)
    {
	ddCoord3DS *pspan;
 	int *pwidth, *pdepth;
	int width, depth;

	pt1 = *ppt++;
	pt2 = *ppt;
	dx = pt2.x - pt1.x;
	dy = pt2.y - pt1.y;
	dz = pt2.z - pt1.z;
	adx = abs(dx);
	ady = abs(dy);
	adz = abs(dz);
	signdx = sign(dx);
	signdy = sign(dy);
	signdz = sign(dz);

	dw = adz;
	if (adx > ady)
	{
	    du = adx;
	    dv = ady;
	    len = adx;
	}
	else
	{
	    du = ady;
	    dv = adx;
	    len = ady;
	}

	e1 = dv * 2;
	e2 = e1 - 2*du;
	e = e1 - du;
	ez1 = dw * 2;
	ez2 = ez1 - 2*du;
	ez = ez1 - du;

	if (ady >= list_len)
	{
	    ddCoord3DS *npspanInit;
	    int *npwidthInit, *npdepthInit;

	    if (local)
	    {
		DEALLOCATE_LOCAL(pdepthInit);
		pdepthInit = (int *)NULL;
		DEALLOCATE_LOCAL(pwidthInit);
		pwidthInit = (int *)NULL;
		DEALLOCATE_LOCAL(pspanInit);
		pspanInit = (ddCoord3DS *)NULL;
		local = FALSE;
	    }
	    list_len = ady + 1;
	    npspanInit = (ddCoord3DS *)xrealloc(pspanInit,
					       sizeof(ddCoord3DS) * list_len);
	    if (!npspanInit)
	    {
		list_len = 0;
		continue;
	    }
	    pspanInit = npspanInit;
	    npwidthInit = (int *)xrealloc(pwidthInit, sizeof(int) * list_len);
	    if (!npwidthInit)
	    {
		list_len = 0;
		continue;
	    }
	    pwidthInit = npwidthInit;
	    npdepthInit = (int *)xrealloc(pwidthInit, sizeof(int) * list_len);
	    if (!npdepthInit)
	    {
		list_len = 0;
		continue;
	    }
	    pdepthInit = npdepthInit;
	}
	pspan = pspanInit;
	pwidth = pwidthInit;
	pdepth = pdepthInit;

	x = pt1.x;
	y = pt1.y;
	z = pt1.z;
	*pspan = pt1;
	if (adx > ady)
	{
	    /* X_AXIS */
	    width = 0;
	    depth = 0;
	    while(len)
	    {
	        depth += signdz;
		if (((signdz > 0) && (ez < 0)) ||
		    ((signdz < 0) && (ez <=0))
		   )
		{
		    ez += ez1;
		    z += signdz;
		}
		else
		{
		    len--;
		    ez += ez2;
		    x += signdx;
		    z += signdz;
		    width++;
		    if (((signdx > 0) && (e < 0)) ||
			((signdx <=0) && (e <=0)))
		    {
		        e += e1;
		    }
		    else
		    {
		        /* give this span a width */
		        *pwidth++ = width;
			*pdepth++ = depth = abs(depth);
			
			/* force the span the right way */
			if (signdx < 0) {
			    pspan->x -= (width-1);
			    pspan->z -= (depth-1);
			}

			/* initialize next span */
			y += signdy;
			e += e2;

			width = 0;
			depth = 0;
			pspan++;
			pspan->x = x;
			pspan->y = y;
			pspan->z = z;
		    }
		}
	    };
	    /* do the last span */
	    *pwidth++ = width;
	    *pdepth++ = depth;
	    if (signdx < 0) {
		pspan->x -= (width-1);
		pspan->z -= (depth-1);
	    }
	}
	else
	{
	    /* Y_AXIS */
	    while(len)
	    {
	        depth += signdz;
		if (((signdz > 0) && (ez < 0)) ||
		    ((signdz < 0) && (ez <=0))
		   )
		{
		    ez +=ez1;
		}
		else
		{
		    len--;
		    z += signdz;
		    ez += ez2;
		    if (((signdx > 0) && (e < 0)) ||
			((signdx <0) && (e <=0))
			)
		    {
		        e +=e1;
		    }
		    else
		    {
		        x += signdx;
			e += e2;
		    }
		    y += signdy;
		    pspan++;
		    pspan->x = x;
		    pspan->y = y;
		    pspan->z = z;
		    *pwidth++ = 1;
		    *pdepth++ = abs(depth);
		    depth = 0;
		}
	    };
	}

	ZBuffer_FillSpans(zbs_ptr, dst, pgc, pwidth-pwidthInit,
			  pspanInit, pwidthInit, pdepthInit);
    }
    if (local)
    {
	DEALLOCATE_LOCAL(pwidthInit);
	DEALLOCATE_LOCAL(pdepthInit);
	DEALLOCATE_LOCAL(pspanInit);
    }
    else
    {
	xfree(pwidthInit);
	xfree(pdepthInit);
	xfree(pspanInit);
    }

    if ((pgc->capStyle != CapNotLast) &&
	((ppt->x != pptInit->x) ||
	 (ppt->y != pptInit->y) ||
	 (ppt == pptInit + 1)))
    {
	int width = 1;
	int depth = 1;
	ZBuffer_FillSpans(zbs_ptr, dst, pgc, 1, ppt, &width, &depth);
    }
} 

miZZeroDashLine(zbs_ptr, dst, pgc, mode, nptInit, pptInit)
ddZBufferStruct *zbs_ptr;
DrawablePtr dst;
GCPtr pgc;
int mode;
int nptInit;		/* number of points in polyline */
ddCoord3DS *pptInit;	/* points in the polyline */
{
    /* XXX kludge until real zero-width dash code is written */
    pgc->lineWidth = 1;
    miZWideDash (zbs_ptr, dst, pgc, mode, nptInit, pptInit);
    pgc->lineWidth = 0;
}
